
#include "funcoes.h"
#include <map>
#include <list>

typedef list<TNo*> TListaNos;
typedef multimap<int, TNo *> TListaNoAEstrela;

TListaNos expande( TNo* );
bool noJaVisitado( TNo* , TListaNos );
bool estadoValido( TEstado , int , int );
void liberaMemoria ( TListaNos lista );
void liberaMemoria ( TListaNoAEstrela lista );

//---------------------------------------------------------------------------
void imprime_resultado(TNo* pt)
{
  if(pt)
  {
    imprime_resultado(pt->getPai());
    cout << "(" << pt -> getEstado().getMissionario() << ", ";
    cout << pt->getEstado().getCanibal() << ", ";
    cout << pt->getEstado().getBarco() <<  ")" << endl;
  }
}
//---------------------------------------------------------------------------
TListaNos busca( string tipoBusca,
                 int profundidadeLimite,
                 int missionarios,
                 int canibais,
                 unsigned int &memoria_maxima,
                 unsigned int &nos_gerados,
                 unsigned int &nos_visitados,
                 unsigned int &nos_memoria,
                 int &nivel_maximo)
{
  TEstado raiz(missionarios,canibais, 1), estadoAtual;
  TNo *pt_no, *pt_noAtual;
  TListaNos arvoreBusca;
  TListaNos expandidos;
  TListaNos visitados;
  int nivel;

  pt_no = new (TNo);

  pt_no -> setNivel( 0 );
  pt_no -> setPai( NULL );
  pt_no -> setEstado( raiz );

  arvoreBusca.push_back( pt_no );

  memoria_maxima = 1;
  nos_gerados = 1;
  nos_visitados = 0;
  nivel_maximo = 0;

  while (!arvoreBusca.empty())
  {
    // Armazena o Valor Mximo de Memoria Utilizada
    if ( (arvoreBusca.size() + visitados.size()) > memoria_maxima)
      memoria_maxima = (arvoreBusca.size() + visitados.size());

    nos_visitados++;
    pt_noAtual = arvoreBusca.front();
    estadoAtual = pt_noAtual -> getEstado();

    nivel = pt_noAtual -> getNivel();

    if (nivel_maximo < nivel) nivel_maximo = nivel;

    arvoreBusca.pop_front();

    if (tipoBusca == "LARGURA")
    {
       visitados.push_back(pt_noAtual);
    }
    else
    {
       // Contornar o Bug da STL
       if (visitados.size() > 0)
          pt_no = visitados.front();
       else
          pt_no = NULL;

       while ( ( (pt_noAtual -> getPai()) != pt_no) && (pt_no != NULL))
       {
         visitados.pop_front();
         free(pt_no);
         pt_no = visitados.front();
       }

       visitados.push_front(pt_noAtual);
    }

    //goal teste
    if ((estadoAtual.getMissionario() == 0) &&
        (estadoAtual.getCanibal() == 0) &&
        (estadoAtual.getBarco() == 0))
    {
      break;
    }

    expandidos = expande(pt_noAtual);

    while (!expandidos.empty())
    {
      pt_no = expandidos.front();
      expandidos.pop_front();

      if (estadoValido( pt_no->getEstado(), missionarios, canibais ))
      {
        if (tipoBusca == "LARGURA")
        {
          arvoreBusca.push_back(pt_no);
          nos_gerados++;
        }
        else if (tipoBusca == "PROFUNDIDADE")
          {
            // Busca em Profundidade pura
            if  ( profundidadeLimite == -1 )
              {
                if ( !noJaVisitado( pt_no, visitados ))
                {
                  nos_gerados++;
                  arvoreBusca.push_front(pt_no);
                }
              }
            else
              {
                if ( (pt_no->getNivel() <= profundidadeLimite) && (nivel != -1) )
                {
                  //no testo ns repetidos, busca iterativa e limitada
                  arvoreBusca.push_front(pt_no);
                  nos_gerados++;
                }
                else
                  free(pt_no);
              }
          }
      }
      else
        free(pt_no);
    }
  }
  nos_memoria = arvoreBusca.size() + visitados.size();
  return visitados;
}
//---------------------------------------------------------------------------
TListaNos AEstrela(int missionarios,
                   int canibais,
                   unsigned int &memoria_maxima,
                   unsigned int &nos_gerados,
                   unsigned int &nos_visitados,
                   unsigned int &nos_memoria,
                   int &nivel_maximo)
{
  TEstado raiz( missionarios, canibais, 1), estadoAtual;
  TNo *pt_no, *pt_noAtual;
  TListaNoAEstrela arvoreBusca;
  TListaNos expandidos;
  TListaNos visitados;
  TListaNoAEstrela::iterator it;
  int nivel, heuristicaPai, heuristicaFilho;

  pt_no = new (TNo);

  pt_no -> setNivel( 0 );
  pt_no -> setPai( NULL );
  pt_no -> setEstado( raiz );

  arvoreBusca.insert( pair< int, TNo*> (pt_no -> getHeuristica(), pt_no) );

  memoria_maxima = 1;
  nos_gerados = 1;
  nos_visitados = 0;
  nivel_maximo = 0;

  while(!arvoreBusca.empty())
  {

    // Armazena o Valor Mximo de Memoria Utilizada
    if ( (arvoreBusca.size() + visitados.size()) > memoria_maxima)
      memoria_maxima = (arvoreBusca.size() + visitados.size());

    nos_visitados++;
    //pega o primeiro da fila
    it = arvoreBusca.begin();

    pt_noAtual = (*it).second;
    estadoAtual = pt_noAtual -> getEstado();
    heuristicaPai = pt_noAtual -> getHeuristica();

    nivel = pt_noAtual -> getNivel();

    if (nivel_maximo < nivel) nivel_maximo = nivel;

    arvoreBusca.erase(it);

    if ((estadoAtual.getMissionario() == 0) &&
        (estadoAtual.getCanibal() == 0) &&
        (estadoAtual.getBarco() == 0))
    {
      visitados.push_back(pt_noAtual);
      break;
    }

    visitados.push_back(pt_noAtual);
    expandidos = expande(pt_noAtual);

    while (!expandidos.empty())
    {
      pt_no = expandidos.front();
      expandidos.pop_front();

      if (estadoValido( pt_no -> getEstado(), missionarios, canibais ))
      {
        if ( !noJaVisitado( pt_no, visitados ))
        {
          //Calcula a heurstica do filho "nmero de missionrios na margem origem"
          heuristicaFilho = (pt_no -> getEstado().getMissionario());

          //Foro a heurstica ser admissvel
          if ( heuristicaFilho < heuristicaPai )
            heuristicaFilho = heuristicaPai;

          arvoreBusca.insert( pair< int, TNo* > ( heuristicaFilho + nivel + 1, pt_no) );

          nos_gerados++;
        }
      }
      else
        free(pt_no);
    }
  }
  nos_memoria = arvoreBusca.size() + visitados.size();
  return visitados;
}
//---------------------------------------------------------------------------
TListaNos expande( TNo* n )
{
  TListaNos l;
  TEstado x;
  TNo * pt_novoNo;
  TOperadores op;


	x = op.mover1m( n -> getEstado() );
  pt_novoNo = new(TNo);
	pt_novoNo -> setPai ( n );
  pt_novoNo -> setNivel ( n -> getNivel() + 1  );
	pt_novoNo -> setEstado ( x );
	l.push_back( pt_novoNo );

	x = op.mover1c( n -> getEstado() );
  pt_novoNo = new(TNo);
	pt_novoNo -> setPai ( n );
  pt_novoNo -> setNivel ( n -> getNivel() + 1  );
	pt_novoNo -> setEstado ( x );
	l.push_back( pt_novoNo );

	x = op.mover2m( n -> getEstado() );
  pt_novoNo = new(TNo);
	pt_novoNo -> setPai ( n );
  pt_novoNo -> setNivel ( n -> getNivel() + 1  );
	pt_novoNo -> setEstado ( x );
	l.push_back( pt_novoNo );

	x = op.mover2c( n -> getEstado() );
  pt_novoNo = new(TNo);
	pt_novoNo -> setPai ( n );
  pt_novoNo -> setNivel ( n -> getNivel() + 1  );
	pt_novoNo -> setEstado ( x );
	l.push_back( pt_novoNo );

	x = op.mover1m1c( n -> getEstado() );
  pt_novoNo = new(TNo);
	pt_novoNo -> setPai ( n );
  pt_novoNo -> setNivel ( n -> getNivel() + 1  );
	pt_novoNo -> setEstado ( x );
	l.push_back( pt_novoNo );

	return l;
}
//---------------------------------------------------------------------------
bool estadoValido( TEstado e, int missionario, int canibal )
{
  int m, c;

  m = e.getMissionario();
  c = e.getCanibal();

  /* Trato os limites Inferiores "menor que 0(zero) e
     Superiores "Qtde Maxima de Missionarios e Canibais" */
  if ((c < 0) || (m < 0) || (c > canibal) || (m > missionario))  return false;

  /* Trato Restricao do Problema: "No de Missionarios >= No de Canibais" */
  if ( ((c > m) && (m != 0)) ||  // margem de origem
       (((canibal - c) > (missionario -m)) && ((missionario - m) != 0))) // margem destino
     return false;

  return true;
}
//---------------------------------------------------------------------------
//Testa se um n j foi visitado
bool noJaVisitado( TNo* no, TListaNos lista)
{
  for( TListaNos::iterator it = lista.begin(); it != lista.end(); it++ )
  {
    if (((*it)->getEstado().getMissionario() == no -> getEstado().getMissionario())&&
       ((*it)->getEstado().getCanibal() == no -> getEstado().getCanibal())&&
       ((*it)->getEstado().getBarco() == no -> getEstado().getBarco()))
    {
      return true;
    }
  }
  return false;
}

//---------------------------------------------------------------------------

